const { performance, PerformanceObserver } = require('node:perf_hooks');
let profiler;
if (process.env.DISABLE_GCPROFILER != "true") profiler = require('gc-profiler');
const si = require('systeminformation');

class AppPerformanceMonitor {
    constructor() {
        this.config = {
            prope_interval: 1000
        }

        this.data = {
            uptime: new Date().getTime(),
            gc_duration: [],
            memory: {
                rss: [],
                heapTotal: [],
                heapUsed: [],
            },
            cpu: {
                system: [],
            },
            process: {
                requests: [],
                handles: [],
            },
            eventloop: {
                utilization: [],
                latency: [],
            }
        }

        // Check if we actualy wanna run this because it leads to binary header issues on some systems
        if (process.env.DISABLE_GCPROFILER != "true") {
            // Start GC Profiler
            profiler.on('gc', (info) => {
                this.data.gc_duration.push(info.duration);
            });
        }

        // Start Performance Observer (Eventloop)
        this.obs = new PerformanceObserver((items) => {
            this.data.eventloop.latency.push(items.getEntries()[0]['duration']);
            this.data.eventloop.utilization.push(items.getEntries()[0]['duration'] / this.config.prope_interval);
            performance.clearMarks();
        });
        this.obs.observe({ entryTypes: ['measure'] });

        // Class loop
        setInterval(this.#loop, this.config.prope_interval);
    }

    /**
     * Data collection loop
     */
    #loop = () => {
        performance.mark('EventLoop_Start'); // Register Eventloop start
        //Process Eventloop delay right after everything within the loop is done
        setImmediate(() => {
            performance.mark('EventLoop_End');
            performance.measure('EventLoop Start -> End', 'EventLoop_Start', 'EventLoop_End');
        });

        // Collect cpu usage

        // Collect memory usage
        const usedHeapSize = process.memoryUsage();
        this.data.memory.rss.push(usedHeapSize.rss);
        this.data.memory.heapTotal.push(usedHeapSize.heapTotal);
        this.data.memory.heapUsed.push(usedHeapSize.heapUsed);

        // Collect other stats
        this.data.process.requests.push(process._getActiveRequests().length);
        this.data.process.handles.push(process._getActiveHandles().length);

        si.currentLoad().then(data => {
            this.data.cpu.system.push(100 - data.currentLoadIdle);
        })
    }

    /**
     * Calculate the average of an array
     * @param {Array} array 
     * @returns 
     */
    #calcAverage = (array) => {
        const sum = array.reduce((a, b) => a + b, 0);
        return (sum / array.length) || 0;
    }


    /**
     * Keeps the passed array at a certain length (30 times prope_interval)
     * @param {Array} array 
     */
    #shiftArrayAndCut = (array) => {
        const length = (this.config.prope_interval / 1000) * 30;
        if (array.length > length) {
            array.shift();
        }
    }

    getStats() {
        const outSystem = {
            uptime: parseInt(((new Date().getTime() - this.data.uptime) / 1000).toFixed(0)),
            gc_duration: this.#calcAverage(this.data.gc_duration),
            memory: {
                rss: this.#calcAverage(this.data.memory.rss),
                heapTotal: this.#calcAverage(this.data.memory.heapTotal),
                heapUsed: this.#calcAverage(this.data.memory.heapUsed),
            },
            cpu: {
                system: this.#calcAverage(this.data.cpu.system),
            },
            process: {
                requests: this.#calcAverage(this.data.process.requests),
                handles: this.#calcAverage(this.data.process.handles),
            },
            eventloop: {
                utilization: this.#calcAverage(this.data.eventloop.utilization),
                latency: this.#calcAverage(this.data.eventloop.latency),
            }
        }

        // Shift arrays
        this.#shiftArrayAndCut(this.data.gc_duration);
        this.#shiftArrayAndCut(this.data.memory.rss);
        this.#shiftArrayAndCut(this.data.memory.heapTotal);
        this.#shiftArrayAndCut(this.data.memory.heapUsed);
        this.#shiftArrayAndCut(this.data.cpu.system);
        this.#shiftArrayAndCut(this.data.process.requests);
        this.#shiftArrayAndCut(this.data.process.handles);
        this.#shiftArrayAndCut(this.data.eventloop.utilization);
        this.#shiftArrayAndCut(this.data.eventloop.latency);

        return outSystem
    }
}

module.exports = AppPerformanceMonitor;